#ifndef __FLOW
#define __FLOW

// This file contains the essential parts to create pointer control flow graphs

#include <stdlib.h>

// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4800 )
// IDA references an internal type as both class and struct..
#pragma warning( disable : 4099 )

#include "legacy.hpp"
#include "graphnode.hpp"
#include "insx86.hpp"
#include "ptrgraph.hpp"
#include "objrec.hpp"
#include "flow.hpp"

//--------------------------------------------------------------------------
//      Creating the code block/access graph

PtrGraph *CreateFlowPtrGraph(ea_t begin, reg_t reg, unsigned int stackofs, rec_t recmode)
{
   PtrGraph       *Flow;
   GraphNode      *root;

   Flow = new(PtrGraph);   // new graph
   root = Flow->GetRootNode();

   AddFlow(Flow, root, begin, reg, stackofs, recmode);
   return(Flow);
}

#define DEBUG
#ifdef DEBUG
   int   DbgFileCount = 0;
   char  DbgFileName[20];
#endif

void  AddFlow(PtrGraph *PtrGraph, GraphNode *from, ea_t begin, reg_t reg, unsigned int stackofs, rec_t recmode)
{
   char           name[30];
   GraphNode      *curr_node;
   ea_t           currinstr = begin, xref_to, xref_from, ea_next_call;
   unsigned int   isize;
   struc_t        *frame;
   member_t       *stack_var, *new_stackvar;
   int            stackdiff;
   area_t         *fnc_area;

   insn_t         cmd_A;

   stack_var = NULL;                    /* keep lint happy */

   sprintf(name, "%08lx-%d-%d", begin, reg, stackofs);
   curr_node = PtrGraph->FindNodeByName(name);

   if(curr_node)                       // A node at this location exists
   {                                   // check if it is for the same pointer/reg
      from->AddDownLink(curr_node); // we have been here before, abort
      return;
   }

   curr_node = PtrGraph->AddNode(from, name);
   if ( curr_node == NULL )
   {
        msg("[!] Memory exhausted! Aborting ...\n");
        return ;
   }
   curr_node->SetBegin(begin);
   curr_node->SetReg(reg);
   curr_node->SetVarOfs(stackofs);

#ifdef DEBUG
/*  if(reg == r_ebp)
   {
      sprintf(DbgFileName, "%08lx.txt", DbgFileCount++);
      PtrGraph->SortByBegin();
      PtrGraph->OutputWingraphFile(DbgFileName);
   }*/
#endif


   if(reg == r_none) // stackvar mode
   {
      frame = get_frame(currinstr);
      if((frame == (struc_t *)BADADDR)||(frame == NULL))
      {
//         msg("[!] No frame found at %lx ! Aborting ...\n", currinstr);
         curr_node->SetEnd(currinstr);
         return;
      }
      stack_var = get_member(frame, stackofs);
      if((stack_var == NULL)||(stack_var == (member_t *)BADADDR))
      {
//         msg("[!] No stackvar found at %lx at offset %lx ! Aborting ...\n", currinstr,  stackofs);
         curr_node->SetEnd(currinstr);
         return;
      }
   }

   while(currinstr != BADADDR)   // the loop which steps through the code
   {
      ua_ana0(currinstr);           // current instruction
      memcpy(&cmd_A, &cmd, sizeof(insn_t));
      isize = cmd_A.size;
      if(isize == 0)
         isize = 1;

      // Let's do all the pointer access crap now ...
      // quite messy code, but I do not know immediately how to clean it
      // up
      if(recmode != rec_justflow)
      {
         if(reg == r_none)
         {  // stack variable code
            if((cmd_A.Op1.reg == r_ebp)||   // only EBP & ESP are valid as regs
               (cmd_A.Op1.reg == r_esp)||
               (cmd_A.Op2.reg == r_ebp)||
               (cmd_A.Op2.reg == r_esp))
            {
               if(get_stkvar((cmd_A.Op1), cmd_A.Op1.addr, &(cmd_A.Op1.addr)) == stack_var)
               {              // stackvar in operand 1
                  if(cmd_A.itype == x86_push)
                  {
//                     msg("[!] push found, need to recurse in __cdecl !\n");
                     if((recmode == rec_cdecl)||(recmode == rec_full))
                     {                                                     /* Darko, 01.05.2003 22:58 */
                        if((ea_next_call = associate_push_to_call(currinstr)) != BADADDR)
                        {
                           stackdiff = get_spd(get_func(currinstr), currinstr);
                           stackdiff -= get_spd(get_func(ea_next_call), ea_next_call);
                           stackdiff += 4; // adjust for return address
                              // we still have to insert code here for dealing with
                              // stub jumps and with infinite looping
                           fnc_area = funcs.get_area(get_next_cref_from(ea_next_call, get_first_cref_from(ea_next_call)));
                           if(fnc_area != 0)
                              new_stackvar = get_member_by_name(get_frame(fnc_area->startEA) , " s");
                           else
                              new_stackvar = 0;

                           if(new_stackvar != 0)
                           {
                              stackdiff += new_stackvar->soff;
                              curr_node->SetEnd(ea_next_call);
                              AddFlow(PtrGraph, curr_node, get_next_cref_from(ea_next_call, get_first_cref_from(ea_next_call)), r_none, stackdiff, recmode);
                              if(reg != r_eax)
                                 AddFlow(PtrGraph, curr_node, get_first_cref_from(ea_next_call), reg, stackofs, recmode);
                           }
                        }
                     }
                  }
                  else if((cmd_A.itype == x86_movA)||(cmd_A.itype == x86_movB))
                  {
//                     msg("[!] stack variable overwritten, aborting at %lx !\n", currinstr);
                     curr_node->SetEnd(currinstr);
                     return;
                  }
                  else if((cmd_A.itype == x86_test)||(cmd_A.itype == x86_cmp))
                  {
//                     msg("[!] test, nevermind\n");
                  }
                  else
                  {
                     msg("[!] unahdnled cmd.itype/stackvar as Op1 at %lx !\n", currinstr);
                     curr_node->SetEnd(currinstr);
                     return;
                  }
               }
               else if(get_stkvar((cmd_A.Op2), cmd_A.Op2.addr, &(cmd_A.Op2.addr)) == stack_var)
               {           // stackvar in Op2
                  if((cmd_A.itype == x86_movA)||(cmd_A.itype == x86_movB))
                  {
                     curr_node->SetEnd(currinstr);
                     AddFlow(PtrGraph, curr_node, currinstr+isize, (reg_t) cmd_A.Op1.reg, 0, recmode);
                     AddFlow(PtrGraph, curr_node, currinstr+isize, r_none, stackofs, recmode);
                     return;
                  }
               }
            }
         }  // done with stack variables
         else if(cmd_A.Op1.reg == reg)
         {  // uarghs, register parser...
            if((cmd_A.itype == x86_push)&&(cmd_A.Op1.type == o_reg))
            {
               curr_node->SetNodeText("pushes & will recurse !\n");
               if((recmode == rec_cdecl)||(recmode == rec_full))
               {                                                 /* Darko,01.05.2003 22:59 */
                  if((ea_next_call = associate_push_to_call(currinstr)) != BADADDR)
                  {
                     stackdiff = get_spd(get_func(currinstr), currinstr);
                     stackdiff -= get_spd(get_func(ea_next_call), ea_next_call);
                     stackdiff += 4; // adjust for return address
                        // we still have to insert code here for dealing with
                        // stub jumps and with infinite looping
                     fnc_area = funcs.get_area(get_next_cref_from(ea_next_call, get_first_cref_from(ea_next_call)));

                     new_stackvar = get_member_by_name(get_frame(fnc_area->startEA) , " s");
                     if(new_stackvar != 0)
                     {
                        stackdiff += new_stackvar->soff;

                        curr_node->SetEnd(ea_next_call);
                        AddFlow(PtrGraph, curr_node, get_next_cref_from(ea_next_call, get_first_cref_from(ea_next_call)), r_none, stackdiff, recmode);
                        if(reg != r_eax)  // remember that eax gets killed in call
                           AddFlow(PtrGraph, curr_node, get_first_cref_from(ea_next_call), reg, stackofs, recmode);
                        return;
                     }
                     else
                     {
//                        msg("[!] could not push-recurse as no target function was found\n");
                     }
                  }
               }
            }
            else if(cmd_A.itype == x86_push)
            {
//               msg("[!] push with offset found at %lx\n", currinstr);
            }
            else if(cmd_A.Op1.type == o_reg)
            {
               // msg("[!] direct register overwrite found, break   \n");
               if((cmd_A.itype != x86_test)&&(cmd_A.itype != x86_cmp))
               {
                  curr_node->SetEnd(currinstr);
                  return;
               }
            }
         }
         else if(cmd_A.Op2.reg == reg)
         {  // oki, Op1 was easy & quick, Op2 is a lot uglier
            if((cmd_A.Op2.type == o_reg)&&((cmd_A.itype == x86_movA)||(cmd_A.itype == x86_movB)))
            {
               if(cmd_A.Op1.type == o_reg)
               {
                  curr_node->SetEnd(currinstr);
                  AddFlow(PtrGraph, curr_node, currinstr+isize, (reg_t)cmd_A.Op1.reg, 0, recmode);
                  AddFlow(PtrGraph, curr_node, currinstr+isize, reg, 0, recmode);
                  return;
               }
               else if((cmd_A.Op1.reg == r_esp)||(cmd_A.Op1.reg == r_ebp))
               {  // move to a stackvar found
                  curr_node->SetEnd(currinstr);
                  AddFlow(PtrGraph, curr_node, currinstr+isize, (reg_t)cmd_A.Op2.reg, 0, recmode);
                  ua_ana0(currinstr);
                  stack_var = get_stkvar((cmd_A.Op1), cmd_A.Op1.addr, &(cmd_A.Op1.addr));
                  if(stack_var != 0)
                  {
//                     msg("move to stackvar found at %lx - stackvar name is %s\n", currinstr, get_member_name(stack_var->id));
                  }

                  if((stack_var != 0)&&(stack_var != (member_t *)BADADDR))
                     AddFlow(PtrGraph, curr_node, currinstr+isize, r_none, stack_var->soff, recmode);
                  return;
               }
            }
         }
      }
      // Do all the control flow checking now

      xref_from = get_first_fcref_from(currinstr);
      if(xref_from != BADADDR)
      {
         // it would be way nicer to base this entire mess solely on
         // cross-references and lastXR, but as ilfak forgot to export
         // lastXR I have to create this mess

         if((cmd_A.itype == x86_jmpA)||(cmd_A.itype == x86_jmpB)||(cmd_A.itype == x86_jmpC)||(cmd_A.itype == x86_jmp_short))
         {  // non-conditional branch ... I WANT LASTXR !
            curr_node->SetEnd(currinstr);
            AddFlow(PtrGraph, curr_node, get_first_fcref_from(currinstr), reg, stackofs, recmode);
            return;
         }
         else if((cmd_A.itype == x86_callA)||(cmd_A.itype == x86_callB)||(cmd_A.itype == x86_callC))
         {  // call ... I WANT LASTXR !
            if((recmode == rec_register)||(recmode == rec_full))
            {
               curr_node->SetEnd(currinstr);    // recurse into subfunctions
               AddFlow(PtrGraph, curr_node, get_first_fcref_from(currinstr), reg, stackofs, recmode);
               AddFlow(PtrGraph, curr_node, get_first_cref_from(currinstr), reg, stackofs, recmode);
            }
         }
         else  // everything else is a conditional branch
         {  // conditional branch ... I WANT LASTXR !
            curr_node->SetEnd(currinstr);
            AddFlow(PtrGraph, curr_node, get_first_fcref_from(currinstr), reg, stackofs, recmode);
            AddFlow(PtrGraph, curr_node, get_first_cref_from(currinstr), reg, stackofs, recmode);
            return;
         }
      }
      if(cmd_A.itype == x86_retn)
      {
         curr_node->SetEnd(currinstr);
         return;
      }
      xref_to = get_first_fcref_to(currinstr+isize);  // next instructions
                                                      // xrefs
      if(xref_to != BADADDR)
      {
         curr_node->SetEnd(currinstr);
         AddFlow(PtrGraph, curr_node, currinstr+isize, reg, stackofs, recmode);
         return;
      }
      currinstr += isize;
   }
}
#endif
